/*
 * Copyright (c) 2008-2010 Lu, Chao-Ming (Tetralet).  All rights reserved.
 * 
 * This file is part of LilyTerm.
 *
 * LilyTerm is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * LilyTerm is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with LilyTerm.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "property.h"

extern struct Command command[COMMAND];

void adjust_ansi_color(GdkColor color[COLOR], GdkColor color_orig[COLOR], gdouble color_brightness)
{
#ifdef DETAIL
	g_debug("! Launch adjust_ansi_color() with color_brightness = %f",  color_brightness);
#endif
	// g_debug("Get win_data = %d when set background saturation!", win_data);
	if (color_brightness > 1) color_brightness = 1;
	if (color_brightness < -1) color_brightness = -1;

	gint i;
	for (i=0; i<COLOR; i++)
	{
		adjust_ansi_color_severally(&color[i], &color_orig[i], color_brightness);
		//g_debug("color_orig[%d] = %x, %x, %x",
		//	i,
		//	color_orig[i].red,
		//	color_orig[i].green,
		//	color_orig[i].blue);

		//g_debug("color[%d] = %x, %x, %x",
		//	i,
		//	color[i].red,
		//	color[i].green,
		//	color[i].blue);
	}
}

void adjust_ansi_color_severally(GdkColor *color, GdkColor *color_orig, gdouble color_brightness)
{
	if (color_brightness>=0)
	{
		(*color).red = (0xffff - (*color_orig).red) * color_brightness + (*color_orig).red;
		(*color).green = (0xffff - (*color_orig).green) * color_brightness + (*color_orig).green;
		(*color).blue = (0xffff - (*color_orig).blue) * color_brightness + (*color_orig).blue;
	}
	else
	{
		(*color).red = (1 + color_brightness) * (*color_orig).red;
		(*color).green = (1 + color_brightness) * (*color_orig).green;
		(*color).blue = (1 + color_brightness) * (*color_orig).blue;
	}
}


// to init a new page
void init_new_page(struct Window *win_data,
		   struct Page *page_data,
		   gint column,
		   gint row,
		   gint run_once)
{
#ifdef DETAIL
	g_debug("! Launch init_new_page() with win_data = %p, page_data = %p, "
		" column = %d, row = %d, run_once = %d",
		win_data, page_data, column, row, run_once);
#endif

	// g_debug("Get win_data = %d when init new page!", win_data);

	// set font
	// g_debug("Set Font AA = %d", win_data->font_anti_alias);
	vte_terminal_set_font_from_string_full (VTE_TERMINAL(page_data->vte),
						page_data->font_name,
						win_data->font_anti_alias);
	//g_debug("Got font size from %s: %d\n", font_name, pango_font_description_get_size (
	//	  pango_font_description_from_string(font_name))/PANGO_SCALE);

	// set terminal size
	// g_debug("Set the vte size to: %dx%d", column, row);
	vte_terminal_set_size(VTE_TERMINAL(page_data->vte), column, row);

	set_vte_color(win_data, page_data);

	// set transparent
	set_background_saturation(NULL, 0, win_data->background_saturation, page_data->vte);

	// other settings
	vte_terminal_set_word_chars(VTE_TERMINAL(page_data->vte), win_data->word_chars);
	vte_terminal_set_scrollback_lines(VTE_TERMINAL(page_data->vte), win_data->scrollback_lines);

	// some fixed parameter
	vte_terminal_set_scroll_on_output(VTE_TERMINAL(page_data->vte), FALSE);
	vte_terminal_set_scroll_on_keystroke(VTE_TERMINAL(page_data->vte), TRUE);
	// vte_terminal_set_backspace_binding (VTE_TERMINAL(page_data->vte), VTE_ERASE_ASCII_DELETE);
	// vte_terminal_set_delete_binding (VTE_TERMINAL(page_data->vte), VTE_ERASE_ASCII_DELETE);
	// vte_terminal_set_emulation (VTE_TERMINAL(page_data->vte), "xterm");

	set_hyprelink(win_data, page_data);

	vte_terminal_set_cursor_blinks (VTE_TERMINAL(page_data->vte), win_data->cursor_blinks);
	vte_terminal_set_audible_bell (VTE_TERMINAL(page_data->vte), win_data->audible_bell);
	vte_terminal_set_visible_bell (VTE_TERMINAL(page_data->vte), win_data->visible_bell);
	vte_terminal_set_backspace_binding (VTE_TERMINAL(page_data->vte), win_data->erase_binding);
	vte_terminal_set_emulation (VTE_TERMINAL(page_data->vte), win_data->emulate_term);
	
	if (run_once)
	{
		win_data->update_hints = 1;
		// update hints and make window unresizable
		window_resizable(win_data->window, page_data->vte, run_once, 1);
	}
}

void set_hyprelink(struct Window *win_data, struct Page *page_data)
{
	if (win_data->enable_hyperlink)
	{
		if (win_data->enable_function_key)
		{
			gint i;
			for (i=0; i<COMMAND; i++)
			{
				page_data->tag[i] = vte_terminal_match_add (VTE_TERMINAL(page_data->vte),
									    command[i].match);
				vte_terminal_match_set_cursor_type(VTE_TERMINAL(page_data->vte),
								   page_data->tag[i],
								   GDK_HAND2);
			}
		}
		else
			vte_terminal_match_clear_all(VTE_TERMINAL(page_data->vte));
	}
}

void set_vte_color(struct Window *win_data, struct Page *page_data)
{
	// set font/background colors
	vte_terminal_set_default_colors(VTE_TERMINAL(page_data->vte));
	if (win_data->using_custom_color)
	{
		// g_debug("win_data->using_custom_color = %d", win_data->using_custom_color);
		vte_terminal_set_colors(VTE_TERMINAL(page_data->vte),
					&(win_data->fg_color),
					&(win_data->bg_color),
					win_data->color, 16);
	}
	else
	{
		vte_terminal_set_color_foreground(VTE_TERMINAL(page_data->vte), &(win_data->fg_color));
		vte_terminal_set_color_background(VTE_TERMINAL(page_data->vte), &(win_data->bg_color));
	}
	vte_terminal_set_color_bold (VTE_TERMINAL(page_data->vte), &(win_data->fg_color));
}

void set_page_width(struct Window *win_data, struct Page *page_data)
{
	if (win_data->fixed_page_width)
		gtk_label_set_width_chars(GTK_LABEL(page_data->label_text), win_data->page_width);
	else
		gtk_label_set_max_width_chars(GTK_LABEL(page_data->label_text), win_data->page_width);
}

void pack_vte_and_scrollbar_to_hbox(struct Window *win_data, struct Page *page_data)
{
	if (win_data->scrollbar_position)
	{
		gtk_box_pack_start(GTK_BOX(page_data->hbox), page_data->vte, TRUE, TRUE, 0);
		gtk_box_pack_start(GTK_BOX(page_data->hbox), page_data->scrollbar, FALSE, FALSE, 0);
	}
	else
	{
		gtk_box_pack_end(GTK_BOX(page_data->hbox), page_data->vte, TRUE, TRUE, 0);
		gtk_box_pack_end(GTK_BOX(page_data->hbox), page_data->scrollbar, FALSE, FALSE, 0);
	}
	if (! win_data->show_scrollbar)
		gtk_widget_set_no_show_all (page_data->scrollbar, TRUE);
}

void hide_scrollbar(struct Window *win_data, struct Page *page_data)
{
	if (win_data->show_scrollbar)
		gtk_widget_show (page_data->scrollbar);
	else
		gtk_widget_hide (page_data->scrollbar);
}

void add_remove_page_timeout_id(struct Window *win_data, struct Page *page_data)
{
	if (page_data->timeout_id)
	{
		g_source_remove (page_data->timeout_id);
		// get_and_update_page_name(page_data);
		// win_data->page_names_no++;
		//if ((page_data->custom_page_name == NULL) && (page_data->label_text != NULL))
		//	 page_data->custom_page_name = g_strdup(page_data->page_name);
		page_data->timeout_id = 0;
	}
	else
		// monitor_cmdline(page_data->monitor, page_data->pid);
		// monitor_cmdline(page_data->channel, page_data->pid);
#ifdef USE_TIMEOUT_SECONDS
		page_data->timeout_id = g_timeout_add_seconds (1,
							       (GSourceFunc)monitor_cmdline,
							       page_data);
#else
		page_data->timeout_id = g_timeout_add (1000,
						       (GSourceFunc)monitor_cmdline,
						       page_data);
#endif
}

void add_remove_window_title_changed_signal(struct Page *page_data)
{
	if (page_data->page_shows_window_title)
		page_data->window_title_signal = g_signal_connect(page_data->vte, "window-title-changed",
								  G_CALLBACK(update_page_window_title), page_data);
	else if (page_data->window_title_signal)
	{
		g_signal_handler_disconnect(page_data->vte, page_data->window_title_signal);
		page_data->window_title_signal = 0;
	}
		
}

gboolean set_background_saturation(GtkRange *range, GtkScrollType scroll, gdouble value, GtkWidget *vte)
{
#ifdef DETAIL
	g_debug("! Launch set_background_saturation() with value = %f, vte = %p", value, vte);
#endif

	struct Page *page_data = (struct Page *)g_object_get_data(G_OBJECT(vte), "Page_Data");
	struct Window *win_data = (struct Window *)g_object_get_data(G_OBJECT(page_data->window), "Win_Data");
	// g_debug("Get win_data = %d when set background saturation!", win_data);

	if (value > 1)
		value =1;
	if (value < 0)
		value =0;

#ifdef ENABLE_RGBA
	if (win_data->use_rgba)
	{
		if (win_data->transparent_background)
			vte_terminal_set_opacity(VTE_TERMINAL(vte), (1-value)*65535);
		else
			vte_terminal_set_opacity(VTE_TERMINAL(vte), 65535);
	}
	else
#endif
		vte_terminal_set_background_transparent(VTE_TERMINAL(vte), win_data->transparent_background);

	// g_debug("win_data->transparent_background = %d, value = %1.3f",
	//	win_data->transparent_background, value);
	if (win_data->transparent_background)
		vte_terminal_set_background_saturation( VTE_TERMINAL(vte), value);
	else
		vte_terminal_set_background_saturation( VTE_TERMINAL(vte), 0);

	return FALSE;
}

#ifdef ENABLE_RGBA
gboolean set_window_opacity(GtkRange *range, GtkScrollType scroll, gdouble value, GtkWidget *window)
{
#ifdef DETAIL
	g_debug("! Launch set_window_opacity() with value = %f, window = %p", value, window);
#endif

	struct Window *win_data = (struct Window *)g_object_get_data(G_OBJECT(window), "Win_Data");
	// g_debug("Get win_data = %d when set window opacity!", win_data);

	if (value > 1)
		value =1;
	if (value < 0)
		value =0;

	if (win_data->use_rgba)
	{
		if (win_data->transparent_window)
			gtk_window_set_opacity (GTK_WINDOW(window), 1-value);
		else
			gtk_window_set_opacity (GTK_WINDOW(window), 1);
	}
	return FALSE;
}
#endif

// set the window hints information
void window_resizable(GtkWidget *window, GtkWidget *vte, gint run_once, gint minsize)
{
#ifdef DETAIL
	g_debug("! Launch window_resizable() with window = %p, vte = %p, run_once = %d, min_size = %d",
		window, vte, run_once, minsize);
#endif
	// vte=NULL when creating a new root window with drag & drop.
	// if (vte==NULL) return;

	GdkGeometry hints = {0};
	// g_debug("Trying to get padding...\n");
	vte_terminal_get_padding (VTE_TERMINAL(vte), &hints.base_width, &hints.base_height);
	
	if (run_once)
	{
		// Only run once when init window or fonts had been changed.
		if (run_once==1)
		{
			hints.width_inc = vte_terminal_get_char_width(VTE_TERMINAL(vte));
			hints.height_inc = vte_terminal_get_char_height(VTE_TERMINAL(vte));
		}
		if (run_once==2)
		{
			hints.width_inc = 1;
			hints.height_inc = 1;
		}
		// g_debug("hints.base_width = %d, hints.base_height = %d\n",
		//	hints.base_width, hints.base_height);
		// g_debug("hints.width_inc = %d, hints.height_inc = %d",
		//	hints.width_inc, hints.height_inc);
	}

	// minsize = -1: the size of vte can NOT be changed.
	if (minsize == -1)
	{
		hints.min_width = minsize;
		hints.min_height = minsize;
	}
	else
	{
		hints.min_width = hints.base_width + hints.width_inc;
		hints.min_height = hints.base_height + hints.height_inc;
	}

	// g_debug("Tring to set geometry on %p, and run_once = %d\n", vte, run_once);
	gtk_window_set_geometry_hints (GTK_WINDOW (window), GTK_WIDGET (vte), &hints,
					GDK_HINT_RESIZE_INC | GDK_HINT_MIN_SIZE | GDK_HINT_BASE_SIZE);

	//g_debug("current the size of vte %p whith hinting = %ld x %ld\n",
	//			vte,
	//			vte_terminal_get_column_count(VTE_TERMINAL(vte)),
	//			vte_terminal_get_row_count(VTE_TERMINAL(vte)));
}

void apply_new_win_data_to_page (struct Window *win_data_orig,
				 struct Window *win_data,
				 struct Page *page_data)
{
//	// 0: do NOT use rgba
//	// 1: force to use rgba
//	// 2: decide by gdk_screen_is_composited()
//	if (win_data_orig->use_rgba != win_data->use_rgba)
//	{
//	}
//#ifdef ENABLE_RGBA
//	if (win_data_orig->transparent_window != win_data->transparent_window)
//	{
//	}
//	if (win_data_orig->window_opacity != win_data->window_opacity)
//	{
//	}
//#endif
	if (win_data_orig->enable_hyperlink != win_data->enable_hyperlink)
		set_hyprelink(win_data, page_data);

// ---- the color used in vte ---- //
	gboolean update_color = FALSE;

	if (compare_color(win_data_orig->fg_color, win_data->fg_color) ||
	    compare_color(win_data_orig->bg_color, win_data->bg_color) ||
	    (win_data_orig->using_custom_color != win_data->using_custom_color) ||
	    (win_data_orig->color_brightness != win_data->color_brightness))
	    	update_color = TRUE;
	
	gint i;
	if ( ! update_color)
	{
		for (i=0; i<COLOR; i++)
			if (compare_color(win_data_orig->color[i], win_data->color[i]))
				update_color = TRUE;
	}
	if (update_color)
		set_vte_color(win_data, page_data);

// ---- tabs on notebook ---- //

	if ((win_data_orig->page_width != win_data->page_width) ||
	    (win_data_orig->fixed_page_width != win_data->fixed_page_width))
		set_page_width(win_data, page_data);

	if ((win_data_orig->page_shows_current_cmdline || win_data_orig->page_shows_current_dir || win_data_orig->page_shows_window_title) !=
	    (win_data->page_shows_current_cmdline || win_data->page_shows_current_dir || win_data->page_shows_window_title))
		add_remove_page_timeout_id(win_data, page_data);
	
	if (win_data->page_shows_window_title != win_data_orig->page_shows_window_title)
		add_remove_window_title_changed_signal(page_data);
	
	// If the new page don't use page_shows_current_cmdline & page_shows_current_dir
	switch (page_data->page_update_method)
	{
		case 0:
			apply_new_page_color (win_data, page_data);
			break;
		case 1:
			if (win_data->page_shows_window_title != win_data_orig->page_shows_window_title)
				apply_new_page_color (win_data, page_data);
			break;
		case 2:
			if (win_data->page_shows_current_cmdline != win_data_orig->page_shows_current_cmdline)
				apply_new_page_color (win_data, page_data);
			break;
		case 3:
			if (win_data->page_shows_current_dir != win_data_orig->page_shows_current_dir)
				apply_new_page_color (win_data, page_data);
			break;
		case 4:
			if (win_data->page_shows_window_title ||
			    win_data->page_shows_current_cmdline ||
			    win_data->page_shows_current_dir)
				apply_new_page_color (win_data, page_data);
			break;
	}

// ---- font ---- //

	if (win_data_orig->font_anti_alias != win_data->font_anti_alias)
		vte_terminal_set_font_from_string_full (VTE_TERMINAL(page_data->vte),
							page_data->font_name,
							win_data->font_anti_alias);

// ---- other settings for init a vte ---- //

	if (strcmp(win_data_orig->word_chars, win_data->word_chars))
		vte_terminal_set_word_chars(VTE_TERMINAL(page_data->vte), win_data->word_chars);

	if (win_data_orig->show_scrollbar != win_data->show_scrollbar)
		hide_scrollbar(win_data, page_data);

	if (win_data_orig->scrollbar_position != win_data->scrollbar_position)
	{
		g_object_ref(page_data->vte);
		gtk_container_remove (GTK_CONTAINER(page_data->hbox), page_data->vte);
		g_object_ref(page_data->scrollbar);
		gtk_container_remove (GTK_CONTAINER(page_data->hbox), page_data->scrollbar);

		pack_vte_and_scrollbar_to_hbox(win_data, page_data);

		g_object_unref(page_data->vte);
		g_object_unref(page_data->scrollbar);
	}

	if ((win_data_orig->transparent_background != win_data->transparent_background) ||
	    (win_data_orig->background_saturation != win_data->background_saturation))
		set_background_saturation (NULL, 0, win_data->background_saturation, page_data->vte);

	// FIXME: The menu_active_window = win_data_orig or win_data here?
	// if ((win_data_orig->use_scrollback_lines != win_data->use_scrollback_lines) ||
	//     (win_data_orig->scrollback_lines != win_data->scrollback_lines))
	// {
	//	win_data->use_scrollback_lines = ! win_data->use_scrollback_lines;
	//	clean_scrollback_lines(NULL, FALSE);
	//}

	if (win_data_orig->cursor_blinks != win_data->cursor_blinks)
		vte_terminal_set_cursor_blinks (VTE_TERMINAL(page_data->vte), win_data->cursor_blinks);

	if (win_data_orig->audible_bell != win_data->audible_bell)
		vte_terminal_set_audible_bell (VTE_TERMINAL(page_data->vte), win_data->audible_bell);

	if (win_data_orig->visible_bell != win_data->visible_bell)
		vte_terminal_set_visible_bell (VTE_TERMINAL(page_data->vte), win_data->visible_bell);
	
	if (win_data_orig->erase_binding != win_data->erase_binding)
		vte_terminal_set_backspace_binding (VTE_TERMINAL(page_data->vte), win_data->erase_binding);
}

void apply_new_page_color(struct Window *win_data, struct Page *page_data)
{
	if (page_data->custom_page_name == NULL)
		page_data->custom_page_name = g_strdup(page_data->page_name);

	if (page_data->is_root)
		page_data->tab_color = win_data->user_page_color[4];
	else    
		page_data->tab_color = win_data->user_page_color[3];
}

// Will return TRUE if a and b are NOT the same.
gboolean compare_color(GdkColor a, GdkColor b)
{
	if ((a.pixel != b.pixel) ||
	    (a.red != b.red) ||
	    (a.green != b.green ) ||
	    (a.blue != b.blue))
		return TRUE;
	else
		return FALSE;
}
